package ssw.mj.codegen;

import static ssw.mj.Errors.Message.NO_VAL;
import static ssw.mj.Errors.Message.NO_VAR;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;

import ssw.mj.Parser;
import ssw.mj.symtab.Tab;

/**
 * MicroJava Code Generator.
 */
public class Code {
	// instruction codes
	public static final int load = 1;
	public static final int load_0 = 2;
	public static final int load_1 = 3;
	public static final int load_2 = 4;
	public static final int load_3 = 5;
	public static final int store = 6;
	public static final int store_0 = 7;
	public static final int store_1 = 8;
	public static final int store_2 = 9;
	public static final int store_3 = 10;
	public static final int getstatic = 11;
	public static final int putstatic = 12;
	public static final int getfield = 13;
	public static final int putfield = 14;
	public static final int const_0 = 15;
	public static final int const_1 = 16;
	public static final int const_2 = 17;
	public static final int const_3 = 18;
	public static final int const_4 = 19;
	public static final int const_5 = 20;
	public static final int const_m1 = 21;
	public static final int const_ = 22;
	public static final int add = 23;
	public static final int sub = 24;
	public static final int mul = 25;
	public static final int div = 26;
	public static final int rem = 27;
	public static final int neg = 28;
	public static final int shl = 29;
	public static final int shr = 30;
	public static final int inc = 31;
	public static final int new_ = 32;
	public static final int newarray = 33;
	public static final int aload = 34;
	public static final int astore = 35;
	public static final int baload = 36;
	public static final int bastore = 37;
	public static final int arraylength = 38;
	public static final int pop = 39;
	public static final int dup = 40;
	public static final int dup2 = 41;
	public static final int jmp = 42;
	public static final int jeq = 43;
	public static final int jne = 44;
	public static final int jlt = 45;
	public static final int jle = 46;
	public static final int jgt = 47;
	public static final int jge = 48;
	public static final int call = 49;
	public static final int return_ = 50;
	public static final int enter = 51;
	public static final int exit = 52;
	public static final int read = 53;
	public static final int print = 54;
	public static final int bread = 55;
	public static final int bprint = 56;
	public static final int trap = 57;

	// codes for compare operators
	public static final int eq = 0;
	public static final int ne = 1;
	public static final int lt = 2;
	public static final int le = 3;
	public static final int gt = 4;
	public static final int ge = 5;

	private static final int[] inverse = { ne, eq, ge, gt, le, lt };

	/** Code buffer */
	public byte[] buf;

	/**
	 * Program counter. Indicates next free byte in code buffer.
	 */
	public int pc;

	/** PC of main method (set by parser). */
	public int mainpc;

	/** Length of static data in words (set by parser). */
	public int dataSize;

	/** According parser. */
	private Parser parser;

	// ----- initialization

	public Code(Parser p) {
		parser = p;
		buf = new byte[100];
		pc = 0;
		mainpc = -1;
		dataSize = 0;
	}

	// ----- code storage management

	public void put(int x) {
		if (pc == buf.length) {
			buf = Arrays.copyOf(buf, buf.length * 2);
		}
		buf[pc++] = (byte) x;
	}

	public void put2(int x) {
		put(x >> 8);
		put(x);
	}

	public void put4(int x) {
		put2(x >> 16);
		put2(x);
	}

	public void put2(int pos, int x) {
		int oldpc = pc;
		pc = pos;
		put2(x);
		pc = oldpc;
	}

	public int get(int pos) {
		return buf[pos];
	}

	public int get2(int pos) {
		return (get(pos) << 8) + (get(pos + 1) & 0xFF);
	}

	// TODO Exercise 5 - 6: implementation of code generation

	/**
	 * Write the code buffer to the output stream.
	 */
	public void write(OutputStream os) throws IOException {
		int codeSize = pc;
		// uncomment for debugging output
		// Decoder.decode(buf, 0, codeSize);
		put('M');
		put('J');
		put4(codeSize);
		put4(dataSize);
		put4(mainpc);
		os.write(buf, codeSize, pc - codeSize); // header
		os.write(buf, 0, codeSize); // code
		os.flush();
		os.close();
	}

	/**
	 * String representation for JUnit test cases.
	 */
	public String dump() {
		StringBuilder sb = new StringBuilder();
		Decoder dec = new Decoder();
		sb.append(dec.decode(buf, 0, pc));
		sb.append("\n#CodeSize: ");
		sb.append(pc);
		sb.append("\n#DataSize: ");
		sb.append(dataSize);
		sb.append("\n#MainPC: ");
		sb.append(mainpc);
		return sb.toString();
	}
}
